#
# $Copyright (c) 2022 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other
# names may be trademarks of their respective owners.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
# $Revision: 1.6.2.18 $
# $Source: /project/arrayrep/src/htcsnap/unix/HTCSnap.pm,v $
#

package HTCSnap;
use ArrayRep;
use ag_i18n_inc;
use File::Basename;
use POSIX qw( uname );
@ISA = ("ArrayRep");

use strict;
use warnings;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default %vmattrib );

@EXPORT = qw(
);
$VERSION = '1.00';


my $VXDG   = "/usr/sbin/vxdg";
my $VXDISK = "/usr/sbin/vxdisk";
my $VXASSIST = "/usr/sbin/vxassist";
my $VXDMPADM = "/opt/VRTS/bin/vxdmpadm";
my $VXCLUSTADM = "/etc/vx/bin/vxclustadm";
my $VXDCTL = "/usr/sbin/vxdctl";
my $DEVNULL = "/dev/null";

if ( $^O =~ /^MSWin/i ){
    my $VMPATH = $ENV{"VMPATH"};
    $VMPATH = "\"" . Win32::GetShortPathName($VMPATH) . "\"";
    $VXDG = $VMPATH . "vxdg";
    $VXDISK = $VMPATH . "vxdisk";
    $VXASSIST = $VMPATH . "vxassist";
    $VXDMPADM = $VMPATH . "vxdmpadm";
    $VXCLUSTADM = "vxclustadm";
    $VXDCTL = "vxdctl";
    $DEVNULL = "nul";
}

my $RECREATEVGTIMER = 15;    # If recreatevg fails,  how many times do we retry
my $RECREATEVGINTVL = 5; # If recreatevg fails, interval to wait before retrying

my ($pairdisplay, $pairsplit, $pairevtwait, $pairresync, $horctakeover, $horcctl);

%default = (
    resname        => '',
    target         => '',
    sotype         => '',
    basedir        => '',
    group          => '',
    dgname         => '',
    instant        => '',
    shadowinstance => '',
    snapdglist     => '',
    snapshotOK     => '',
    tcinstance     => '',
    vmtype         => '',
    RAC            => '',    # = 1 if this FD is being run in a RAC env
	                         # = 0, otherwise
    # Possible values: AIX/Solaris/Linux/HPUX
    OStype         => '',
    user           => '',
    domain         => '',
    enc_pword      => '',
);

%vmattrib = (
    LVMVG     => 'VolumeGroup',
    DiskGroup => 'DiskGroup',
    CVMVolDg  => 'CVMDiskGroup',
    VMDg      => 'DiskGroupName',
);

sub new {
    my ( $proto, $resname, $target, $mountSO, $useSO, $requireSO,
		$shadowinstance, $snapdglist, $snapshotOK, $user, $domain, $enc_pword ) = @_;

    my ( $group, $basedir, $instance, $dgname );
    my $class = ref($proto) || $proto;
    my @dgname;

    my $self = {%default};
    $self->{'resname'}        = $resname;
    $self->{'target'}         = $target;
    $self->{'shadowinstance'} = $shadowinstance;
    $self->{'snapdglist'}     = $snapdglist;
    $self->{'snapshotOK'}     = $snapshotOK;
    $self->{'user'}           = $user;
    $self->{'domain'}         = $domain;
    $self->{'enc_pword'}      = $enc_pword;

    # Possible values: AIX/HP-UX/Linux/SunOS/Windows NT
    $self->{'OStype'} = (uname())[0];
    chomp $self->{'OStype'};

	if ($mountSO) {
		if ($requireSO) {
			$self->{'sotype'} = "gold-only";
		} else {
			$self->{'sotype'} = 'gold';
		}
	} elsif ($useSO) {
		if ($requireSO) {
			$self->{'sotype'} = "silver-only";
		} else {
			$self->{'sotype'} = 'silver';
		}
	} else {
		$self->{'sotype'} = "bronze";
	}

	$self->{'instant'} = 0;

    if ( $^O !~ /^MSWin/i ) {
        ##################### BEGIN CHANGES FOR RAC FD ##################
        $self->{'RAC'} = 0;
        my ($sg)  = split( '\n', `$ag_i18n_inc::HARES -value $target Group` );
        my ($pll) = split( '\n', `$ag_i18n_inc::HAGRP -value $sg Parallel` );
        if ( $pll == 1 ) {
            $self->{'RAC'} = 1;
        }
        ###################### END CHANGES FOR RAC FD ###################
    }

	# get DG name
	my ($typ) = split( '\n', `$ag_i18n_inc::HARES -value $target Type` );
	my $index = 0;
	my @tmp   = "";
	my $dgattrib = "";
	if ( $typ eq "HTC" ) {
		my (@depfields) = split( '\n', `$ag_i18n_inc::HARES -dep $target` );
		foreach my $depline (@depfields) {
			my (@deppair) = split( '\s+', $depline );
			if ( $deppair[2] eq $target ) {
				@tmp =
				  split( '\n', `$ag_i18n_inc::HARES -value $deppair[1] Type` );
				if ( 0 == length $self->{'vmtype'} ) {
					$self->{'vmtype'} = $tmp[0];
					$dgattrib = $vmattrib{ $self->{'vmtype'} };
				} else {
					if ( $self->{'vmtype'} ne $tmp[0] ) {
						VCSAG_LOG_MSG( "E",
							"Mixed VM types are not supported; exiting", 7801 );
						return undef;
					}
				}

				@tmp = split( '\n',
					`$ag_i18n_inc::HARES -value $deppair[1] $dgattrib` );
				$dgname[$index] = $tmp[0];
				$index = $index + 1;
			}
		}
		$self->{'dgname'} = join( ' ', @dgname );
		($group) = split( '\n', `$ag_i18n_inc::HARES -value $target GroupName` );
		($instance) =
		  split( '\n', `$ag_i18n_inc::HARES -value $target Instance` );
		($basedir) = split( '\n', `$ag_i18n_inc::HARES -value $target BaseDir` );
	} else {
		my ($sg) = split( '\n', `$ag_i18n_inc::HARES -value $target Group` );
		my (@htcres) =
		  split( '\n', `$ag_i18n_inc::HARES -list Type=HTC Group=$sg` );
        if (! (defined $htcres[0])) {
            VCSAG_LOG_MSG( "E", "HTC resource not found. Please verify value of TargetResName attribute",7840 );
            exit(0);
        }
        my (@htcresn) = split( '\s+', $htcres[0] );
		($group) =
		  split( '\n', `$ag_i18n_inc::HARES -value $htcresn[0] GroupName` );
		($instance) =
		  split( '\n', `$ag_i18n_inc::HARES -value $htcresn[0] Instance` );
		($basedir) =
		  split( '\n', `$ag_i18n_inc::HARES -value $htcresn[0] BaseDir` );
		$self->{'vmtype'} = $typ;
        @tmp = split('\n',
            `$ag_i18n_inc::HARES -value $target $vmattrib{$self->{'vmtype'}}`);
        $self->{'dgname'} = $tmp[0];
        $self->{'instant'} = 1;
	}

    if (($self->{'vmtype'} eq "LVMVG") && ($self->{'OStype'} !~ /AIX/i)) {
        VCSAG_LOG_MSG("E", "LVM is supported only on AIX; Exiting", 7802);
        return undef;
    }

	#$self->{'dgname'} = $dgname;
	$self->{'tcinstance'} = $instance;
	$self->{'group'}      = $group;
	$self->{'basedir'}    = $basedir;

	#strip leading spaces if any
	$self->{'dgname'} =~ s/^[\s]+//;

    if ( $^O =~ /^MSWin/i ) {
        $self->{'basedir'} = Win32::GetShortPathName($self->{'basedir'});
        $pairdisplay = "\"$self->{'basedir'}\\pairdisplay\"";
        $pairsplit = "\"$self->{'basedir'}\\pairsplit\"";
        $pairevtwait = "\"$self->{'basedir'}\\pairevtwait\"";
        $pairresync = "\"$self->{'basedir'}\\pairresync\"";
        $horctakeover = "\"$self->{'basedir'}\\horctakeover\"";
        $horcctl = "\"$self->{'basedir'}\\horcctl\"";
    } else {
        $pairdisplay = "$self->{'basedir'}/pairdisplay";
        $pairsplit = "$self->{'basedir'}/pairsplit";
        $pairevtwait = "$self->{'basedir'}/pairevtwait";
        $pairresync = "$self->{'basedir'}/pairresync";
        $horctakeover = "$self->{'basedir'}/horctakeover";
        $horcctl = "$self->{'basedir'}/horcctl";
    }

	bless( $self, $class );
	return $self;
}

sub DESTROY { }

sub take_snapshot {
	my $self              = shift;
	my ($repSplitMode)    = shift;
	my ($shadowSplitMode) = shift;

	my $is_rac_cluster = $self->{'RAC'};
	my $is_master      = 0;
	my $cmd;

    if ( $^O !~ /^MSWin/i ) {
        if ( $is_rac_cluster == 1 ) {

            # This is a RAC environment i.e. the group for which a
            # FireDrill is being run is a parallel group.
            $is_master = ArrayRep::is_cvm_master();

            if ( $is_master == -1 ) {

                # An appropriate error message would
                # already have been logged in the
                # is_master sub
                return 1;
            } elsif ( $is_master == 0 ) {
                # Slave node
                # Dont do anything
                return 0;
            }
        }
    }

	#
	# IMPORTANT: Control should come here only if:
	# This is a non-RAC cluster
	# Or
	# This is a RAC cluster and this system is the CVM Master
	#

	my $ret = 0;
	my ($oto) =
	  split( '\n', `$ag_i18n_inc::HATYPE -value HTCSnap OnlineTimeout` );
	my $count = $oto / 10;

	my $instant = 0;    # consistent multi-LUN snapshots available
	my $noread  = 0;    # no-read snapshots available

	#---------------------------------------------------
	# Verify that the "noread" mode is appropriately set.
	#---------------------------------------------------

	$ENV{'HORCMINST'}  = $self->{'shadowinstance'};
	$ENV{'HORCC_MRCF'} = 1;
	my (@lines) = split( '\n',`$pairdisplay -g $self->{'group'} -CLI -fd` );
	foreach my $l (@lines) {
		my (@fields) = split( '\s+', $l );
		if ( $fields[2] eq "R" ) {
			if ( $fields[11] ne "-" ) {
				# ShadowImage S-Vol has not been altered.
				$self->{'instant'} = 1;
			}
			if ( $fields[11] eq "N" ) {
				$noread = 1;
			}
		}
	}

	if ( ( !$noread ) && ( $self->{'sotype'} =~ /^silver/ ) ) {
		if ( $self->{'sotype'} =~ /silver-only/ ) {
			VCSAG_LOG_MSG( "E",
				"ShadowImage must be set up in noread mode for silver snapshot; exiting",
				7803 );
		} else {
			VCSAG_LOG_MSG( "W",
				"ShadowImage must be set up in noread mode for silver snapshot; using bronze instead",
				7804 );
		}
		return 1;
	}

	if ( ($noread) && ( $self->{'sotype'} =~ /^gold/ ) ) {
		if ( $self->{'sotype'} =~ /gold-only/ ) {
			VCSAG_LOG_MSG( "E",
				"ShadowImage must not be set up in noread mode for gold snapshot; exiting",
				7805 );
		} else {
			VCSAG_LOG_MSG( "W",
				"ShadowImage must not be set up in noread mode for gold snapshot; using bronze instead",
				7806 );
		}
		return 1;
	}

	#---------------------------------------------------
	# Verify that ShadowImage SVols are synced
	#---------------------------------------------------

    $ENV{'HORCMINST'} = $self->{'shadowinstance'};
    $ENV{'HORCC_MRCF'} = 1;
    $cmd =  $pairevtwait ." -g " . $self->{'group'}. " -s pair -t 5";
    $ret = system("$cmd >$DEVNULL");
    if ($ret) {
		VCSAG_LOG_MSG( "E",
			"pre-requisites for taking snapshot not met; exiting", 7807 );
		return $ret;
	}

	if (   ( ( $self->{'sotype'} =~ /^gold/ ) && ( !$self->{'instant'} ) )
		|| ( $self->{'sotype'} =~ /^silver/ ) )
	{

		# pause replication
		# insert commands to pause replication
        $ENV{'HORCMINST'} = $self->{'tcinstance'};
        delete $ENV{'HORCC_MRCF'};
        $cmd =  $pairevtwait . " -g " . $self->{'group'} . " -s pair -t 5";
        $ret = system("$cmd >$DEVNULL");
        if ($ret) {
            VCSAG_LOG_MSG( "E","pairevtwait failed; returning", 7808 );
            return $ret;
        }

        $cmd = $pairsplit . " -g " . $self->{'group'}. " -$repSplitMode";
        $ret = system("$cmd >$DEVNULL");
        if ($ret) {
            VCSAG_LOG_MSG( "E",
                "cannot pause replication in preparation for split; exiting", 7809 );
            return $ret;
        }
	}

	#---------------------------------------------------
	# take snapshots
	# insert commands to snapshot
	#---------------------------------------------------
    $ENV{'HORCMINST'} = $self->{'shadowinstance'};
    $ENV{'HORCC_MRCF'} = 1;
    $cmd = $pairsplit . " -g ". $self->{'group'}. " -$shadowSplitMode";
    $ret = system("$cmd >$DEVNULL");

    if ( $self->{'sotype'} =~ /^gold/ ) {
        $cmd =  $pairevtwait . " -g " . $self->{'group'}. " -s psus -t 5";
        $ret = system("$cmd >$DEVNULL");
        if ($ret) {
            VCSAG_LOG_MSG( "E",
                "pairevtwait for psus failed ; exiting", 7810 );
            return $ret;
        }
    }

	# if snapshot succeeded set temporary attribute to local node name
	# if snapshot failed and requireSO, exit with error
	if ($ret) {
		VCSAG_LOG_MSG( "E", "cannot take required snapshot; exiting", 7811 );
        $ENV{'HORCMINST'} = $self->{'tcinstance'};
        delete $ENV{'HORCC_MRCF'};
        $cmd = "$pairresync -g $self->{'group'}";
        my $retVal = system("$cmd >$DEVNULL");
		if ($retVal) {
			VCSAG_LOG_MSG( "E", "cannot resynchronize after split; exiting",
				7812 );
			return $retVal;
		}
		return $ret;
	} else {
		my (@fields) = split( '\s+', `$ag_i18n_inc::HASYS -nodeid` );
		system(
"$ag_i18n_inc::HARES -modify $self->{'resname'} Responsibility $fields[5]"
		);
		$self->{'snapshotOK'} = 1;
	}

	if ( ( $self->{'sotype'} =~ /^gold/ ) && ( !$self->{'instant'} ) ) {

		# resume replication
		# insert commands to resume replication
        $ENV{'HORCMINST'} = $self->{'tcinstance'};
        delete $ENV{'HORCC_MRCF'};
        $cmd = "$pairresync -g $self->{'group'}";
        $ret = system("$cmd >$DEVNULL");
		if ($ret) {
			VCSAG_LOG_MSG( "E", "cannot resynchronize after split; exiting",
				7813 );
			return $ret;
		}
	}

	return $ret;
}

sub split_replication {
    my $self = shift;

    my $is_rac_cluster = $self->{'RAC'};
    my $is_master      = 0;

    if ( $^O !~ /^MSWin/i ){
        if ( $is_rac_cluster == 1 ) {

            # This is a RAC environment i.e. the group for which a
            # FireDrill is being run is a parallel group.
            $is_master = ArrayRep::is_cvm_master();
            if ( $is_master == -1 ) {

                # An appropriate error message would
                # already have been logged in the
                # is_master sub
                return 1;
            } elsif ( $is_master == 0 ) {

                # Slave node
                # Dont do anything
                return 0;
            }
        }
    }

    #
    # IMPORTANT: Control should come here only if:
    # This is a non-RAC cluster
    # Or
    # This is a RAC cluster and this system is the CVM Master
    #

    my $ret = 0;
    my ($oto) =
        split( '\n', `$ag_i18n_inc::HATYPE -value HTCSnap OnlineTimeout` );
    my $count = $oto / 10;

    # if cannot split, exit with error
    $ENV{'HORCMINST'} = $self->{'tcinstance'};
    delete $ENV{'HORCC_MRCF'};
    my $cmd = $pairevtwait . " -g ". $self->{'group'}. " -s pair -t 5";
    $ret = system("$cmd >$DEVNULL");
    if ($ret) {
        VCSAG_LOG_MSG( "E","pairevtwait failed; returning", 7814 );
        return $ret;
    }
    $cmd = $pairsplit . " -g ". $self->{'group'} . " -rw";
    $ret = system("$cmd >$DEVNULL");
    if ($ret) {
        VCSAG_LOG_MSG( "E", "cannot split replication; exiting", 7815 );
        exit 0;
    }

    return $ret;
}

# For vxvm as the underlying volume manager, the subroutine will clear
# import information on the disks in the device group, update the udid
# of the disks (if it is a rac environment), and set an appropriate tag
# on the disks so that when the vxdg import cmd is fired with the
# appropriate cli options, it will succeed.
# Returns: 0 on success
#          1 on failure
#
sub htc_prepare_disks {
    my ($self)    = shift;
    my ($case)    = shift;
    my ($tagname) = shift;    # We should be able to return the
	                          # tagname value to the caller this way
	                          # tagname is a pass-by-reference argument
    my $ret       = 0;
    my (@disks);
    my ($is_rac_cluster) = $self->{'RAC'};

    my ($searchdev) = "R";
    my $tag = "SIRemote-";

    if ( $case eq "silver" ) {
        $searchdev = "L";
        $tag       = "SILocal-";
    }
    if ( $case eq "bronze" ) {
        $searchdev = "L";
        $tag       = "SILocal-";
    }

    if ( $^O =~ /^MSWin/i ){
        VCSAG_LOG_MSG( "N", "issuing command 'vxassist rescan' after h/w snapshot", 7816);
        $ret = system("$VXASSIST rescan");
        if ($ret) {
            VCSAG_LOG_MSG( "E", "vxassist rescan the devices failed after h/w snapshot", 7817);
        }
    }

    if ( $case ne "bronze" ) {
        # First get the ShadowImage SVOL device names
        $ENV{'HORCMINST'}  = $self->{'shadowinstance'};
        $ENV{'HORCC_MRCF'} = 1;
        my (@lines) = split( '\n',
            `$pairdisplay -g $self->{'group'} -CLI -fd` );

        foreach my $l (@lines) {
            my (@fields) = split( '\s+', $l );
            if ( $fields[2] ne $searchdev ) { next; }
            my $snapdev = $fields[3];

            if ( $^O !~ /^MSWin/i ) {
                if (($self->{'vmtype'} ne "LVMVG") && ($self->{'OStype'} =~ /AIX/i)) {
                    # A raw device has an r as the first letter in
                    # the device name. We need to skip that.
                    # We may have to do the same for Linux as well.
                    # Need to verify this though.
                    $snapdev =~ s/^r//;
                }

                # Map the physical dev name to the vm dev name
                my (@devnames) = split('\n', `$VXDMPADM getdmpnode nodename=$snapdev`);
                $ret = $?;

                if ($ret) {
                    # Failed to get the dmpnodename of the device
                    VCSAG_LOG_MSG("E",
                        "Failed to get the dmpnodename for device $snapdev, exiting",
                        7818, "$snapdev");
                    return 1;
                }

                foreach my $dname (@devnames) {
                    if (($dname =~ /^NAME/) || ($dname =~ /^====/)) {
                        next;
                    }
                    (@fields) = split('\s+', $dname);
                    $snapdev = $fields[0];
                   last;
                }
            }
            push( @disks, $snapdev );
        }
    }

	# TBD unique tagname across clusters
	$$tagname = $tag . $self->{'group'} . $self->{'shadowinstance'};
	my $final_tag = $tag . $self->{'group'} . $self->{'shadowinstance'};

	# In the future, there should be code to check the type of
	# volume manager technology before calling the appropriate
	# volume-manager-specific routine to set up the disks for an
	# import operation.

    if ( $^O =~ /^MSWin/i ) {
        my (%vxdisk);
        my (@dlines) = split( '\n', `$VXDISK list` );
        foreach my $l (@dlines) {
            my (@fields) = split( '\s+', $l );
            $vxdisk{ $fields[1] } = $fields[0];
        }

        foreach my $disk (@disks) {
            if ( $vxdisk{$disk} ) {
                $disk = $vxdisk{$disk};
            }
            $ret = system("$VXDISK set tag value=$final_tag $disk");
            if ($ret) {
                VCSAG_LOG_MSG( "E", "cannot tag disk $disk; exiting",
                    7819, "$disk" );
                return 1;
            }
        }
    }
    else {
        $ret = ArrayRep::vxvm_prepare_disks( $final_tag, $is_rac_cluster, @disks );
    }

    return $ret;
}

#
# sub mount_snapshot {
# Renaming the subroutine from mount_snapshot to mount_diskgroup since a
# lot of the functionality is common between all cases of firedrill, but
# we only mount the snapshot in the gold case of firedrill. So the name
# mount_snapshot for the subroutine could be misleading.
#
# Arguments to the subroutine:
# ----------------------------
# self: Resource instance
# case: Type of firedrill: gold/silver/bronze
#
# Return Values:
# --------------
# 0 on success
# 1 on failure
#
# Functionality:
# --------------
# The subroutine sets up everything for a successful import of the
# diskgroup on the right set of disks: BCV-luns if gold firedrill and
# DEV-luns in case of silver and bronze firedrills.
# Further, if CVM is the underlying volume manager being used, the
# subroutine will make sure that the disks setup and dg import happens
# only on the CVM MASTER.
#
sub mount_diskgroup {
    my ( $self, $case ) = @_;
    my $ret            = 0;
    my $is_master      = 0;
    my $is_rac_cluster = $self->{'RAC'};

    if ( $^O !~ /^MSWin/i ) {
        if ( $is_rac_cluster == 1 ) {

            # This is a RAC environment i.e. the group for which a
            # FireDrill is being run is a parallel group.
            $is_master = ArrayRep::is_cvm_master();

            if ( $is_master == -1 ) {

                # An appropriate error message would
                # already have been logged in the
                # is_master sub
                return 1;
            } elsif ( $is_master == 0 ) {

                # This node is the SLAVE
                # Wait in a loop until the MASTER
                # imports the shared dg

                $ret = ArrayRep::cvm_slave_wait_for_shared_dg_import(
                    $self->{'dgname'} );
                if ($ret) {

                    # An appropriate error message would
                    # already have been logged in the
                    # An appropriate error message would
                    # already have been logged in the
                    # slave_wait_for_shared_dg_import sub
                    return 1;
                } elsif ( $ret == 0 ) {

                    # Dg has been imported by the
                    # MASTER - so all shd work ok
                    return 0;
                }
            }
        }
    }

    #
    # COME HERE ONLY IF THE NODE IS THE CVM MASTER OR THIS IS A
    # NON-RAC ENVIRONMENT
    #

    # Clear any import info, update udid, set tag on the appropriate
    # disks. Basically setup the disks for the impending dg import
    my $tagname = "";

    # tagname will be passed by reference into the prepare_disks sub
    $ret = $self->htc_prepare_disks( $case, \$tagname );

    if ($ret) {

        # An appropriate error message would
        # already have been logged in the
        # sub prepare_disks
        return 1;
    }

    # Test import of the dg
    # In the future, there should be code to check the type of
    # volume manager technology before calling the appropriate
    # volume-manager-specific routine to test that the import
    # operation succeeds.

    if ( $^O =~ /^MSWin/i ) {
        my @dgname = split( ' ', $self->{'dgname'} );
        my $newdg;
        foreach my $dg (@dgname) {
            $newdg = $self->FireDrillDGName($dg);
            my (@lines) = split( '\n', `$VXDG -C -n $newdg -o useclonedev=on -o updateid -o tag=$tagname import -g $dg`);

            $ret = $?;
            if ($ret) {
				# If vxdg import fails, retry with clean flag
				# Applicable only IS 731 onwards.
				@lines = split( '\n', `$VXDG -C -n $newdg -o useclonedev=on -o updateid -o tag=$tagname import -g $dg -clean`);
				$ret = $?;
			}
			
			if ($ret) {
                VCSAG_LOG_MSG( "E",
                    "cannot import disk group $dg on snapshot; exiting",
                    7820, "$dg" );
                return 1;
            }
        }
    }
    else {
        $ret = ArrayRep::vxvm_test_diskgroup_import( $self->{'dgname'}, $case,
            $tagname, $is_rac_cluster );
    }

    return $ret;
}

sub pitc_gold {
    my ($self) = @_;
    my $ret = 0;

    $ret = $self->take_snapshot( "r", "C" );
    if ($ret) {
        if ( $self->{'sotype'} eq "gold-only" ) {
            VCSAG_LOG_MSG( "E",
                "cannot take required snapshot of group $self->{'group'}; exiting",
                7821, "$self->{'group'}" );
            return $ret;
        } else {
            VCSAG_LOG_MSG( "W",
                "cannot take required snapshot of group $self->{'group'}; dropping to bronze type",
                7822, "$self->{'group'}" );
            $self->pitc_bronze();
        }
    } else {
        my (@fields) = split( '\s+', `$ag_i18n_inc::HASYS -nodeid` );
        system( "$ag_i18n_inc::HARES -modify $self->{'resname'} Responsibility $fields[5]" );
    }

    if (( $self->{'vmtype'} eq "DiskGroup" ) || ( $self->{'vmtype'} eq "CVMVolDg" ) || ( $self->{'vmtype'} eq "VMDg" )) {
        $ret = $self->mount_diskgroup("gold");
    } elsif ( $self->{'vmtype'} eq "LVMVG" ) {
        $ret = $self->mount_snapshot_lvm("gold");
    }

    return $ret;
}

sub pitc_silver {
    my ($self) = @_;
    my $ret = 0;

    $ret = $self->take_snapshot( "rw", "E" );
    if ($ret) {
        if ( $self->{'sotype'} eq "silver-only" ) {
            VCSAG_LOG_MSG( "E",
                "cannot take required snapshot of group $self->{'group'}; exiting",
                7823, "$self->{'group'}" );
            return 1;
        } else {
            VCSAG_LOG_MSG( "W",
                "cannot take required snapshot of group $self->{'group'}; dropping to bronze type",
                7824, "$self->{'group'}" );
            $self->pitc_bronze();
        }
    }

    # Mount the TrueCopy SVol
    if (( $self->{'vmtype'} eq "DiskGroup" ) || ( $self->{'vmtype'} eq "CVMVolDg" ) || ( $self->{'vmtype'} eq "VMDg" )) {
        $ret = $self->mount_diskgroup("silver");
    } elsif ( $self->{'vmtype'} eq "LVMVG" ) {
        VCSAG_LOG_MSG( "E", "Silver is currently not supported with LVM ", 7825 );
        return 1;
    }

    return $ret;
}

sub pitc_bronze {
    my ($self) = @_;
    my @dgname = split( ' ', $self->{'dgname'} );

    VCSAG_LOG_MSG( "N",
        "\n\nBRONZE:Performing read-write replication split\n\n", 7826 );

    # perform read-write replication split
    # insert command to split replication and write enable target
    my $ret = $self->split_replication();
    if ($ret) {
        VCSAG_LOG_MSG( "E",
            "Could not Split the replication in pitc_bronze for $self->{'group'};",
            7827, $self->{'group'} );
        return 1;
    }

    if (( $self->{'vmtype'} eq "DiskGroup" ) || ( $self->{'vmtype'} eq "CVMVolDg" )  || ( $self->{'vmtype'} eq "VMDg" )) {
        $ret = $self->mount_diskgroup("bronze");
    } elsif ( $self->{'vmtype'} eq "LVMVG" ) {
        VCSAG_LOG_MSG( "E", "Bronze is currently not supported with LVM ",
            7828 );
        return 1;
    }

    if (!$ret) {
        VCSAG_LOG_MSG( "N",
            "\n\nBRONZE:Processing completed for group $self->{'group'}\n\n",
            7829, "$self->{'group'}" );

        # touch lock file and exit
        $self->create_lockfile();
    }

    return $ret;
}

sub pitc {
    my ($self) = @_;
    my $ret = 0;
    if ( $self->{'sotype'} =~ /gold/ ) {
        $ret = $self->pitc_gold();
    } elsif ( $self->{'sotype'} =~ /silver/ ) {
        $ret = $self->pitc_silver();
    } else {
        $ret = $self->pitc_bronze();
    }
    return $ret;
}

sub resynch_gold {
    my ($self) = @_;
    my $ret    = 0;

    my ($oto) = split( '\n', `$ag_i18n_inc::HATYPE -value HTC OfflineTimeout` );
    my $count = $oto / 10;

    if ( $self->{'vmtype'} eq "LVMVG" ) {
        my @dgs = map ( $_ . "_fd", ( split( /\s+/, $self->{'dgname'} ) ) );
        $ret = ArrayRep::aix_lvm_exportvgs( \@dgs );
    }

    #TBD consistent snapshot required
    if ( !$self->{'instant'} ) {
        my ($took) = split( '\n',
            `$ag_i18n_inc::HARES -value $self->{'resname'} Responsibility` );
        if ( !$took ) {
            $self->resynch_bronze();
            return 0;
        }
    }

    $ENV{'HORCMINST'}  = $self->{'shadowinstance'};
    $ENV{'HORCC_MRCF'} = 1;
    my $cmd = $pairresync . " -g ". $self->{'group'};
    $ret = system("$cmd >$DEVNULL");
    if ($ret) {
        VCSAG_LOG_MSG(
            "E",
            "cannot resynch required snapshot of group $self->{'group'}; exiting",
            7830,
            "$self->{'group'}"
        );
        return 1;
    } else {
        my (@fields) = split( '\s+', `$ag_i18n_inc::HASYS -nodeid` );
        system(
            "$ag_i18n_inc::HARES -modify $self->{'resname'} Responsibility \"\""
        );
    }

    return 0;
}

sub resynch_silver {
    my ($self) = @_;
    my $ret = 0;

    my (@fields) = split( '\s+', `$ag_i18n_inc::HASYS -nodeid` );
    my ($took) = split( '\n',
        `$ag_i18n_inc::HARES -value $self->{'resname'} Responsibility` );
    if ( !$took ) {

        # Drop through case
        $self->resynch_bronze();
        return 0;
    }
    if ( $took && ( $took eq $fields[5] ) ) {

        $ENV{'HORCMINST'}  = $self->{'tcinstance'};
        delete $ENV{'HORCC_MRCF'};

        my $cmd = $pairresync . " -g ". $self->{'group'};
        $ret = system("$cmd >$DEVNULL");
        if ($ret) {
            VCSAG_LOG_MSG( "E",
                "cannot resynchronize replication; administrative intervention required",
                7831 );
            exit 0;
        }
        my ($oto) =
            split( '\n', `$ag_i18n_inc::HATYPE -value HTCSnap OfflineTimeout` );
        if ( $oto > 60 ) {
            $oto = $oto - 30;
            $ENV{'HORCMINST'}  = $self->{'tcinstance'};
            delete $ENV{'HORCC_MRCF'};

            $cmd = $pairevtwait . " -g " . $self->{'group'} . " -s pair -t $oto";
            $ret = system("$cmd >$DEVNULL");
            if ( !$ret ) {
                $ENV{'HORCMINST'}  = $self->{'shadowinstance'};
                $ENV{'HORCC_MRCF'} = 1;
                $cmd = $pairresync . " -g ". $self->{'group'};
                $ret = system("$cmd >$DEVNULL");
                if ($ret) {
                    VCSAG_LOG_MSG( "E",
                        "cannot resynchronize snapshot; administrative intervention required",
                        7832 );
                    exit 0;
                } else {
                    system(
                        "$ag_i18n_inc::HARES -modify $self->{'resname'} Responsibility \"\""
                    );
                }
            }
        }
        VCSAG_LOG_MSG( "N",
"replication resynchronization requires more time than OfflineTimeout; monitor entry point will resynchronize snapshot once replication resynchronization is complete",
        7833 );
    }
    return 0;
}

sub resynch_bronze {
    my ($self) = @_;

    $ENV{'HORCMINST'}  = $self->{'tcinstance'};
    delete $ENV{'HORCC_MRCF'};

    my $cmd = $pairresync . " -g ". $self->{'group'};
    my $ret = system("$cmd >$DEVNULL");
    if ($ret) {
        VCSAG_LOG_MSG( "E",
            "cannot resynchronize replication; administrative intervention required",
            7834 );
        exit 0;
    }
    return 0;
}

sub resynch {
    my ($self) = @_;
    my $ret = 0;

    if ( $self->{'sotype'} =~ /gold/ ) {
        $ret = $self->resynch_gold();
    } elsif ( $self->{'sotype'} =~ /silver/ ) {
        $ret = $self->resynch_silver();
    } else {
        $ret = $self->resynch_bronze();
    }
    return $ret;
}

sub resynchronize {
    my ($self) = @_;

    my (@fields) = split( '\s+', `$ag_i18n_inc::HASYS -nodeid` );
    my ($took) = split( '\n',
        `$ag_i18n_inc::HARES -value $self->{'resname'} Responsibility` );
    if ( $took && ( $took eq $fields[5] ) ) {
        $ENV{'HORCMINST'}  = $self->{'tcinstance'};
        delete $ENV{'HORCC_MRCF'};
        my $cmd = $pairevtwait . " -g " . $self->{'group'} . " -s pair";
        my $ret = system("$cmd >$DEVNULL");
        if ( !$ret ) {
            $ENV{'HORCMINST'}  = $self->{'shadowinstance'};
            $ENV{'HORCC_MRCF'} = 1;
            $cmd = $pairresync . " -g " . $self->{'group'};
            $ret = system("$cmd >$DEVNULL");
            if ($ret) {
                VCSAG_LOG_MSG( "E",
                "cannot resynchronize snapshot; administrative intervention required",
                7835 );
            }
            system(
                "$ag_i18n_inc::HARES -modify $self->{'resname'} Responsibility \"\""
            );
        } else {
            $ENV{'HORCMINST'}  = $self->{'shadowinstance'};
            $ENV{'HORCC_MRCF'} = 1;
            $cmd = $pairevtwait . " -g " . $self->{'group'} . " -s pair -t 5";
            $ret = system("$cmd >$DEVNULL");
            if ( !$ret ) {
                system(
                    "$ag_i18n_inc::HARES -modify $self->{'resname'} Responsibility \"\""
                );
            }
        }
    }
    return 0;
}

sub start_raidmgr {
    my ($self) = @_;
    my $ret = 0;

    $ENV{'HORCMINST'}  = $self->{'shadowinstance'};
    $ENV{'HORCC_MRCF'} = 1;

    if ( $^O =~ /^MSWin/i ){
        if ( defined $self->{'user'} && $self->{'user'} ne "") {
            system("callcmd.exe \"$self->{'basedir'}\\horcmstart $self->{'shadowinstance'}\" $self->{'user'} $self->{'domain'} $self->{'enc_pword'}");
        } else {
            system("callcmd.exe \"$self->{'basedir'}\\horcmstart $self->{'shadowinstance'}\"");
        }
    } else {
        system("$self->{'basedir'}/horcmstart.sh $self->{'shadowinstance'} > $DEVNULL 2>&1");
    }
    $ret = system("$horcctl -g $self->{'group'} -ND > $DEVNULL 2>&1");
    return ($ret);
}

sub call_run_cmds {
	shift;
	my ($args) = shift;
	my @cmds = ();
	(@cmds) = @$args;
	my $ret = ArrayRep::ar_run_cmds( \@cmds );
	return $ret;
}

sub mount_snapshot_lvm {
	my ( $self, $case ) = @_;
	my $ret = 0;
	my (@SIdisks);
	my ( $vgname, $disk );
	my ($searchdev) = "R";
	my @cmds = ();

	# Map VG names to VG ids
	my %vgs_to_id = ();

	# Map VGid to all its BCV disks
	my %vgid_to_disks = ();

	my $RECREATEVG = "/usr/sbin/recreatevg";
	my $CHDEV      = "/usr/sbin/chdev";

##   if ($case eq "silver") {
##      $searchdev = "L";
##   }

	# First get the ShadowImage SVOL device names
	$ENV{'HORCMINST'}  = $self->{'shadowinstance'};
	$ENV{'HORCC_MRCF'} = 1;
	my (@lines) = split( '\n',
		`$pairdisplay -g $self->{'group'} -CLI -fd` );

	foreach my $l (@lines) {
		my (@fields) = split( '\s+', $l );
		if ( $fields[2] ne $searchdev ) { next; }
		push( @SIdisks, $fields[3] );
	}

	my @vgnames = split( ' ', $self->{'dgname'} );

	# Construct a hash to map VGname to VG serial id
	ArrayRep::aix_lvm_get_vgs_id( \@vgnames, \%vgs_to_id );

	my @vgids = ( values %vgs_to_id );

	# Construct a hash to map VGids to SI disks
	ArrayRep::aix_lvm_get_vgdisks( \@vgids, \%vgid_to_disks, \@SIdisks );

	foreach $vgname ( keys %vgs_to_id ) {
		my $Lv_Prefix = "fd_";

		my $fdvgname = $vgname . "_fd";
		my @disks    = @{ $vgid_to_disks{ $vgs_to_id{$vgname} } };

		foreach $disk (@disks) {
			push( @cmds, "$CHDEV -l $disk -a pv=clear" );
			$ret = ArrayRep::ar_run_cmds( \@cmds );

			if ($ret) {
				VCSAG_LOG_MSG( "W",
					"Could not clear PV id on $disk. recreatevg might fail ",
					7836, "$disk" );
			}
		}

		# At times, recreatevg fails when called immediately after a split.
		# So, sleep for  $RECREATEVGINTVL seconds before trying recreatevg
###      push( @CMDS, "sleep $RECREATEVGINTVL " );
###      run_cmds(@CMDS);

###      push( @CMDS, "$RECREATEVG -y $fdvgname -Y fd_ -n @disks" );
###      $ret = run_cmds(@CMDS);

		for ( my $i = $RECREATEVGTIMER ; $i > 0 ; $i-- ) {
			push( @cmds, "$RECREATEVG -y $fdvgname -Y $Lv_Prefix -n @disks" );
			$ret       = ArrayRep::ar_run_cmds( \@cmds );
			$Lv_Prefix = "NA";
			last if ( !$ret );
			VCSAG_LOG_MSG( "I",
				"recreatevg failed; retrying again after $RECREATEVGINTVL seconds; $i more retries left",
				7837, "$RECREATEVGINTVL", "$i" );
			push( @cmds, "sleep $RECREATEVGINTVL " );
			ArrayRep::ar_run_cmds( \@cmds );
		}

		if ($ret) {
			VCSAG_LOG_MSG( "E", "Cannot recreate VG $fdvgname; exiting",
				7838, "$fdvgname" );
			return 1;
		}

		my $line = `$ag_i18n_inc::HARES -list VolumeGroup=$fdvgname`;
		my ( $lvmresname, undef ) = split( /\s+/, $line );

		ArrayRep::aix_lvm_update_majornum_attrib( $lvmresname, \@disks );
	}

	return 0;
}

sub reenable_vxconfigd_on_slaves {
    my $self = shift;
    my $is_rac_cluster = $self->{'rac'};
    my $is_master = 0;
    my (@v1) = ();
    my $nodename = "";
    my $ret = 0;
    my $reenable_failed = 0;

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($par) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    if ($par != 1) {
        return 0;
    }
    my ($sup) = split('\n', `$ag_i18n_inc::HATYPE -value CVMVolDg SupportedActions`);
    if (! ($sup =~ /vxdctlenable/) ) {
        return 0;
    }

    if ($is_rac_cluster == 1) {
        # RAC cluster - slave does nothing here
        $is_master = ArrayRep::is_cvm_master();

        if ($is_master == -1) {
            return 1;
        }
        elsif ($is_master == 0) {
            # Slave node: dont do anything
            return 0;
        }
    }

    # get the list of all nodes in the cvm cluster
    my (@vxclus) = split('\n', `$VXCLUSTADM nidmap 2>$DEVNULL`);
    foreach my $v (@vxclus) {
        if (($v !~ /Master/) && ($v !~ /^Name/)) {
            # for all slave nodes
            (@v1) = split('\s+', $v);
            $nodename = $v1[0];
            $ret = $self->reenable_vxconfigd($nodename);
            if ($ret == 1) {
                $reenable_failed = 1;
            }
        }
    }
    if ($reenable_failed == 1) {
        VCSAG_LOG_MSG("W","The action to reenable the volume configuration daemon failed on atleast one slave node in the CVM cluster. This may cause the CVM diskgroup to not be imported successfully.", 7839);
    }
}

sub reenable_vxconfigd {
    my ($self, $nodename) = @_;
    my $reenable_failed = 0;

    # get the resources that depend on the htcsnap resource
    my (@deps) = split('\n', `$ag_i18n_inc::HARES -dep $self->{'resname'}`);

    foreach my $d (@deps) {
        if ($d =~ /#Group/) {
            next;
        }
        my (@fields) = split('\s+', $d);
        if ($fields[2] ne $self->{'resname'}) {
            next;
        }
        my ($typ) = split('\n', `$ag_i18n_inc::HARES -value $fields[1] Type`);
        if ($typ eq "CVMVolDg") {
            # Invoke the action to run vxdctl enable on the slave
            # node. It is sufficient to invoke the action for just
            # one CVMVolDg resource on a node, if the action
            # succeeds
            `$ag_i18n_inc::HARES -action $fields[1] vxdctlenable -sys $nodename`;
            $reenable_failed = $?;
            if (!$reenable_failed) {
                # The action succeeded
                last;
            }
        }
    }
    return $reenable_failed;
}

sub FireDrillDGName {

    # Name of diskgroup used by application group. This is input
    my ( $self, $currentDGName ) = @_;

    # Name of the firedrill diskgroup.
    my $firedrillDGName;
    my $standardFDName = $currentDGName . "_fd";

    if ( $self->{'snapdglist'} eq ""){
        # either the values are not provided or attribute is not present
        # return standard name
        # should we log this.
        return $standardFDName;
    }

    # the format is in DGName;FD_DGName and then separated by space.
    # sanjay, we can change format!
    my (@dg_fddg_NameMap) = split(' ', $self->{'snapdglist'});
    foreach my $dg_fddg_NameEntry (@dg_fddg_NameMap) {
        #check till the ;, so that we do not search the portion of the name
        my (@fdDiskGroupName) = split(';', $dg_fddg_NameEntry);
        if ( $fdDiskGroupName[0] eq $currentDGName){
            # found it.
            $firedrillDGName = $fdDiskGroupName[1];
            return $firedrillDGName;
        }
    }
    # well, nothing worked so fall back to existing convention.
    return $standardFDName;
}

sub get_lockfilename {
    my ($self) = shift;
    my ($typ) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Type`);

    my ($lock_file, $lock_dir);

    if ( $^O =~ /^MSWin/i ){
        $lock_dir = $ag_i18n_inc::VCSLOG . "\\lock\\" . $typ . "\\";
    } else {
        $lock_dir = $ag_i18n_inc::VCSLOG . "/lock/" . $typ . "/";
    }

    $lock_file = $lock_dir . ".VCS_" . $typ . "_" . $self->{'resname'} . "_Online_Lockfile.DO-NOT-REMOVE";
    return $lock_file;
}

sub create_lockfile {
    my ($self) = shift;
    my ($lock_file) = $self->get_lockfilename();

    my (undef, $lock_dir, undef) = fileparse($lock_file);

    if (! -d $lock_dir) {
        mkdir  $lock_dir or
            VCSAG_LOG_MSG("E", "Could not create lock directory $lock_dir",
                    7841, "$lock_dir");
    }

    open(LFH, ">$lock_file") or
        VCSAG_LOG_MSG("E", "Could not create file: $lock_file", 7842,
                "$lock_file");

    close(LFH);
}


sub remove_lockfile {
    my($self) = shift;
    my ($lock_file) = $self->get_lockfilename();

    if (-f "$lock_file") {
        unlink("$lock_file") or
            VCSAG_LOG_MSG("E", "Could not delete file: $lock_file", 7843,
                    "$lock_file");
    }
}

#
# The subroutine returns 110 if the online lock file for the resource
# exists on the system; 100 otherwise. The monitor entry point that
# invokes this subroutine must exit with the *return code* from this
# entry point.
#
sub check_lockfile {
    my ($self) = shift;
    my $res_state = 100;
    my ($lock_file) = $self->get_lockfilename();

    if (-f "$lock_file") {
        $res_state = 110;
    }
    return $res_state;
}

1;
